#!/usr/bin/env groovy

import org.xdai.seq.*;
import org.xdai.tool.*;
import groovy.text.*;

def appname="PlantTFcat";

def prompt='''./PlantTFcat.groovy -q <Query Sequence File> [--nodefile=] [--minslicesize=] [--deltmp] [--sessionid=1234567]
''';
def cli = new CliBuilder(usage: prompt);
cli.h(longOpt: 'help', 'print help information');
cli.b(longOpt: 'base', 'Base path of pipeline', args: 1);
cli.q(longOpt: 'query', 'Query sequence file full path', args: 1);
cli.E(longOpt: 'expect', 'Expect value of complementarity, default: '+SessionProps.expect, args: 1);
cli.remoteaddr('remote user IP address, just for recording, optional', args: 1);
cli.nodefile('Node configration file for BioGrid, pipeline will run multi-threads within single computer without it', args: 1);
cli.minslicesize('Minimum slice size of Biogrid, default will be automatically determined by BioGrid', args: 1);
cli.deltmp('Delete user-submitted tmp file from  web server.')
cli.sessionid('unique session ID, can be specified by user',args: 1);
//the following options are email related
cli.email('We will remind you by email upon completion',args: 1);
cli.hostname('hostname of web server',args: 1);
cli.port('port of web server',args: 1);
cli.emailsender('Sender email address',args: 1);
cli.emailreceiver('Reply email address',args: 1);
cli.smtp('smtp address',args: 1);
//end

def opt = cli.parse(args);
if(args.length<4||!opt||opt.h) { cli.usage(); System.exit(-1); }

def base;   if(opt.b) base=opt.b;      if(base==null)   { println 'Please set -b'; System.exit(-1); }
def query;  if(opt.q) query=opt.q;     if(query==null)  { println 'Please set -q'; System.exit(-1); }
def expect=SessionProps.expect;        if(opt.E) expect=opt.E;
def remoteaddr;                        if(opt.remoteaddr)  remoteaddr=opt.remoteaddr;
def nodefile=null;                     if(opt.nodefile)    nodefile=opt.nodefile;
def minslicesize=null;                 if(opt.minslicesize)  minslicesize=opt.minslicesize;
def deltmp=false;                      if(opt.deltmp) deltmp=true;
def sessionid=GeneralTool.getUniqSessionID();  if(opt.sessionid) sessionid=opt.sessionid;  //def sessionid=java.util.UUID.randomUUID().toString();
//email related :
def email=null;                        if(opt.email)  email=opt.email;
def hostname=null;                     if(opt.hostname) hostname=opt.hostname;
def port=null;                         if(opt.port)     port=opt.port;
def emailsender=null;                  if(opt.emailsender) emailsender=opt.emailsender;
def emailreceiver=null;                if(opt.emailreceiver) emailreceiver=opt.emailreceiver;
def smtp=null;                         if(opt.smtp)       smtp=opt.smtp;
if( email!=null && (hostname==null||port==null||emailsender==null||emailreceiver==null||smtp==null) ){
  println "Once you specify email, you also need set hostname, port,smtp, emailsender and emailreceiver.\n";
  System.exit(1);
}

if(!new File(query).exists()){
  println "${query} doesn't exists.\n";
  System.exit(1);
}

base=(base=~/\/+$/).replaceAll("");
def bindir=base+"/bin" as String;
def datadir=base+"/data" as String;
def sessionbase=base+"/session" as String;
def sessiondir=sessionbase+"/"+sessionid as String;
def iprdir=base+"/miniIPR" as String;

def tmpworkdir=new File(sessiondir);
if(tmpworkdir.exists()) tmpworkdir.deleteDir();
tmpworkdir.mkdirs();    //&makeDeleteScript($nodefilename,$sessionworkdir,"delete.sh");
def syncdir="${bindir}:${iprdir}" as String;

//save properties of session
def propsFileFullPath=sessiondir+"/"+SessionProps.propsFile;
def propsFile=new File(propsFileFullPath);
propsFile.append("timestart="+System.currentTimeMillis()+"\n");
propsFile.append("base=${base}\n");
propsFile.append("sessionbase=${sessionbase}\n");
propsFile.append("sessiondir=${sessiondir}\n");
propsFile.append("query=${query}\n");
if(remoteaddr!=null) propsFile.append("remoteaddr=${remoteaddr}\n");
if(nodefile!=null) propsFile.append("nodefile=${nodefile}\n"); 
propsFile.append("expect=${expect}\n");
def binding = ["appname":appname, "sessionid":sessionid, "hostname":hostname, "port":port ];  //bind to text temlate for email content

System.err.println "sessionid=${sessionid}";
if(email!=null){  //++++++++++++++++++++++++++++++++++++++++++++++
   def engine = new GStringTemplateEngine();
   def emailcontent = engine.createTemplate( new File("${bindir}/sessionstart.email.txt") ).make(binding).toString();
   def mailutils = new MailSender();
   mailutils.sendMail(email, emailreceiver, "Your session [${sessionid}] has been submitted to ${appname}", emailcontent, "text/html", "");
}


//**************************Check input*********************************************
System.err.println "Stage 1/3: Pre-checking submitted sequences......";
def checked_query=treatSeqFile(query,sessiondir, bindir);

propsFile.append("checkedQueryFullPath=${checked_query}\n");
def tmpfilesize=(new File(checked_query)).length();
propsFile.append("checkedQueryFileSize=${tmpfilesize}\n");

/********calculate the optimal slicing option for BioGrid************/
//def kilo=1024;  def million=kilo * kilo;
if(nodefile==null) nodefile="localhost.ini";
nodefile=bindir+"/"+nodefile;
def slicefactor=5;
if(minslicesize==null) minslicesize=50;
if(tmpfilesize<1024*1024) minslicesize=10;
if(tmpfilesize<0.1*1024*1024) minslicesize=5;

//**********************Running BioGrid*****************************************
System.err.println "Stage 2/3: Prediction pipeline is analyzing data......";
//ipr.pl args array and command full string: bin/ipr.pl --ipr /root/project/PlantTFcat/miniIPR --query /root/project/PlantTFcat/tmp/pep.fas 
String[] iprArgs=["--ipr", iprdir, "--query", checked_query]; 
def commandString="${bindir}/ipr.sh "+iprArgs.join(" ");
//biogrid Maincall's args array
def iprResultFullPath=sessiondir+"/"+SessionProps.iprTxtResult;
def biogridcmd=["-s", syncdir, "-d", "-p", "4", "-f", slicefactor, "-m", minslicesize, "-u", "200", "-o", iprResultFullPath];  //[-m 1], minimum slice size 
if(nodefile!=null)  { biogridcmd.add("-n");  biogridcmd.add(nodefile); }
biogridcmd.add(commandString);
def biogridArray=biogridcmd as String[];
//call BioGrid ......
try{
   org.xdai.biogrid.Maincall.main(biogridArray);
}catch(Exception e){ System.err.println("ERROR: "+e.getMessage());  }
if(! new File(iprResultFullPath).exists()) { 
   System.err.println("ERROR: Fatal errors happened, BioGrid failed."); 
   System.err.flush(); 
   System.exit(-1); 
}
propsFile.append("iprTxtResult="+SessionProps.iprTxtResult+"\n");
if(email!=null){  //++++++++++++++++++++++++++++++++++++++++
   def engine = new GStringTemplateEngine();
   def emailcontent = engine.createTemplate( new File("${bindir}/sessioncomplete.email.txt") ).make(binding).toString();
   def mailutils = new MailSender();
   mailutils.sendMail(email, emailreceiver, "Your ${appname} session [${sessionid}] has been completed", emailcontent, "text/html", "");
}


//*********************writing database
System.err.println "Stage 3/3: Populating database for web search......";
def tfTxtResultFullPath=sessiondir+"/"+SessionProps.tfTxtResult;
def dbResultFullPath=sessiondir+"/"+SessionProps.dbResult;
try{
    System.err.println "Loading text-based InterProScan data into SQLite......";
    TfAnalysis.txt2db(iprResultFullPath, dbResultFullPath, checked_query);
    System.err.println "Calculating score of TF/TR candidates based on SQLite db and family patterns ......";
    TfAnalysis.predictTFScore("${datadir}/tfdomain.txt" as String, dbResultFullPath, tfTxtResultFullPath, TfAnalysis.loadFamilyDetail(base));
}catch(Exception e) {
    e.printStackTrace();
    System.err.println("ERROR: "+e.getMessage());
    System.err.flush();
    System.exit(-1); 
}
propsFile.append("dbResult="+SessionProps.dbResult+"\n");
propsFile.append("tfTxtResult="+SessionProps.tfTxtResult+"\n");

//***************************finished***********************
propsFile.append("timeend="+System.currentTimeMillis()+"\n");
println "propsFile=${propsFileFullPath}";       //session property file
println "iprTxtResult=${iprResultFullPath}";     //output of InterProScan
println "dbResult=${dbResultFullPath}"           //transcriptioon factor prediction in MySQL format
println "tfTxtResult=${tfTxtResultFullPath}";    //transcriptioon factor predicted by pipeline, tab-delimited

if(deltmp) new File(query).delete();
/*-----THE END------*/


//*************************function region*****************************************
def String treatSeqFile(String seqfile, String sessiondir, String bindir){
  if(!new File(seqfile).exists()) {
	    System.err.println("ERROR: can't found original sequence file, '${seqfile}' doesn't exist");
	    System.exit(-1);
  }
  if(new File(seqfile).length()>SessionProps.seqFileUpLimit) {
	    System.err.println("ERROR: uploaded sequence file is larger than uplimit (${SessionProps.seqFileUpLimit} nt).");
	    System.exit(-1);
  }
  def submitted_seqfile=sessiondir+"/submitted_seqfile";
  System.err.println("Copy user-submitted sequence file into session folder");
  if(!GeneralTool.copyfile(seqfile,submitted_seqfile)){
	    System.err.println("ERROR: failed to copy ${seqfile} to ${submitted_seqfile}");
	    System.exit(-1);
  }
  
  "${bindir}/dos2unix ${submitted_seqfile}".execute().waitFor();
  "${bindir}/mac2unix ${submitted_seqfile}".execute().waitFor();

  //check if submission content is non-fasta format
  if(!StreamFasta.isFASTA(submitted_seqfile, Seq.PROTEIN)){    // Seq.PROTEIN works for both protein and NT 
      if(StreamFasta.isPureSeq(submitted_seqfile, true, Seq.PROTEIN)){
      	 System.err.println("User-submitted sequence file is in PureSeq format, converting to FASTA format.");
      	 try{ StreamFasta.pureSeq2fasta(submitted_seqfile, submitted_seqfile+".fas", SessionProps.seqfileMaxseqlen, Seq.PROTEIN); }  
      	 catch(Exception e) { System.err.println("ERROR: "+e.getMessage()); System.exit(-1); }
      	 submitted_seqfile=submitted_seqfile+".fas";
      }else if(StreamFasta.isBlank(submitted_seqfile)){
      	 System.err.println("ERROR: User-submitted content is blank.");
      	 System.exit(-1);
      }else{
      	 System.err.println("ERROR: User-submitted sequences are neither FASTA nor PureSeq format.");
      	 System.exit(-1);
      }
  }
  
  //translate if submitting NT sequences
  if(StreamFasta.getSeqType(submitted_seqfile) == Seq.NT) {
      def proc = "${bindir}/fastatranslate ${submitted_seqfile}".execute();
      def newsubmittedfn="${submitted_seqfile}.pep";
      def newsubmitted=new File(newsubmittedfn);
      proc.in.eachLine{
             it=(it=~/^>\s*(\S+)(.*):\[revcomp\]:\[translate\(([1-3])\)\]/).replaceAll('>$1_ORF-$3$2');    ////:[revcomp]:[translate(3)]
             it=(it=~/^>\s*(\S+)(.*):\[translate\(([1-3])\)\]/).replaceAll('>$1_ORF+$3$2');  //:[translate(1)]
             newsubmitted.append(it+"\n");
      }
      if(proc.exitValue()!=0){
	 System.err.println("ERROR: failed to translate NT sequences into Protein sequences, program returned: ${proc.err.text}");
	 System.exit(-1);
      }
      submitted_seqfile=newsubmittedfn;
  }


  System.err.println("Check if user-submitted sequence file is well-formatted FASTA file."); 
  def checked_seqfile=submitted_seqfile+'.checked';
  FastaChecker check = new FastaChecker();
  check.setMinseqlen(SessionProps.seqfileMinseqlen);         check.setAbortWhenMinseqlen(false);
  check.setMaxseqlen(SessionProps.seqfileMaxseqlen);         check.setAbortWhenMaxseqlen(false);
  check.setMinnumberseq(SessionProps.seqfileMinnumberseq);   check.setAbortWhenMinnumberseq(true);
  check.setMaxnumberseq(SessionProps.seqfileMaxnumberseq);   check.setAbortWhenMaxnumberseq(false);
  check.setCaseselect(FastaChecker.UPPERCASE);
  check.setValidSeqLen(35);
  check.setValidSeqRatio(0.1);
  BufferedReader bf = null;
  PrintStream    ps = null;
  try{
     bf = new BufferedReader(new InputStreamReader(new FileInputStream(submitted_seqfile)));
     ps = new PrintStream(checked_seqfile);
     check.check(bf, ps, null, Seq.PROTEIN, true);
  }catch(Exception e){
     System.err.println("ERROR: found errors from submitted sequence data: "+e.getMessage()+".");		
     System.exit(-1);
  }finally{
     bf.close();
     ps.close();
  }

  if(StreamFasta.getSeqType(checked_seqfile)!=Seq.PROTEIN){
  	System.err.println("ERROR: submitted sequences aren't valid protein sequences");
  	System.exit(-1);
  }

  def tmpfasta=new Fasta(checked_seqfile);
  tmpfasta.close();
  return checked_seqfile;
}
