import groovy.sql.Sql
import org.xdai.seq.*;



//TfAnalysis2.predictTFScore("/root/project/PlantTFcat/data/tfdomain.txt", "/root/project/PlantTFcat/session/1339434137780162/tf.sqlite","");

public class TfAnalysis {

 public static def loadFamilyDetail(def base){
   def dbfile="${base}/data/families.sqlite";
   def sql = Sql.newInstance("jdbc:sqlite:"+dbfile, "", "", "org.sqlite.JDBC");
   def families=[:];
   sql.rows("select * from tffamily").each{
       families[it.family_acc]=it;
   };
   return families;
 }

 public static def predictTFScore(def tfdomainfile, def dbfile, def tfTxtResultFullPath, def families){
   def seq_ipr_tb=SessionProps.seq_ipr_tb+"2";
   def seq_iprs_tb=SessionProps.seq_iprs_tb;
   def tf_tb=SessionProps.tf_tb;

   //generate family domain pattern hashmap
   def hMusthave=[:];
   def hExclude=[:];
   new File(tfdomainfile).eachLine{
      if(it=~/^\s*$/) return;
      def cols=it.split("\t");
      if(cols.length!=3){
         System.err.println("Found error in ${tfdomainfile}, line: ${it}" as String);
         return;
      }
      hMusthave[cols[0]]=(cols[1] as String).replaceAll("\\s+","").replaceAll(/^[&\|]*|[&\|]*$/,"").replaceAll(/\|+/,"|").replaceAll(/&+/,"&");
      hExclude[cols[0]]=(cols[2] as String).replaceAll("\\s+","").replaceAll(/^[&\|]*|[&\|]*$/,"").replaceAll(/\|+/,"|").replaceAll(/&+/,"&");
   }
   System.err.println "Compiled patterns for "+hMusthave.size()+" families";

   //prepare table ${tf_tb} to host final result
   def sql = Sql.newInstance("jdbc:sqlite:"+dbfile, "", "", "org.sqlite.JDBC");
   sql.execute("create table if not exists ${tf_tb} (family text key, family_type text key, seq_acc text, seq_realacc text key, ipr_accs text, seq_comment text, all_text text key, myscore real, rank int, ID int)" as String);
   sql.execute("PRAGMA journal_mode=OFF");
   sql.execute("PRAGMA synchronous=OFF");
   sql.execute("BEGIN TRANSACTION");


   //calcuting socre for each sequences
   def id=0;
   sql.eachRow("select distinct seq_acc from ${seq_ipr_tb}" as String){
      def tmp_seq_acc = it.seq_acc;   def iprscore=[:];
      sql.eachRow("select ipr_acc,myscore from ${seq_ipr_tb} where seq_acc='${tmp_seq_acc}'" as String, { iprscore[it.ipr_acc]=it.myscore; }); 
      //generate a ipr vs score table
      hMusthave.each{  //scoring each family for the currrent sequence (tmp_seq_acc)
         def tmpfamily=it.key;  def tmpmusthave=it.value; def tmpexclude=hExclude[tmpfamily];
         iprscore.each{  tmpmusthave = tmpmusthave.replaceAll(it.key, it.value as String);  }   //replace IPR domain with corresponding miniIPR score
         iprscore.each{  tmpexclude  = tmpexclude.replaceAll(it.key, it.value as String);   }   //same as above
         tmpmusthave = tmpmusthave.replaceAll(/IPR[T0-9]+/, "0.0"); 
         tmpexclude = tmpexclude.replaceAll(/IPR[T0-9]+/, "0.0");
         def tmpmusthavescore=0.0;
         def tmpexcludescore=0.0;
         try{ 
             tmpmusthavescore=calPatternScore(tmpmusthave); 
             tmpexcludescore=calPatternScore(tmpexclude);
         }catch(Exception e){ 
             System.err.println("Found error for the pattern of ${tmpfamily}/${tmp_seq_acc}:    ${tmpmusthave}   ${tmpexclude}");
         }

         if(tmpmusthavescore==0.0||tmpexcludescore>0.0) return;
         if(tmpmusthavescore>1.0) tmpmusthavescore=1.0;
         def allcols=[tmpfamily, families[tmpfamily].family_type, tmp_seq_acc, '', '', '', '',tmpmusthavescore,1,++id];
         sql.execute("insert into ${tf_tb} values(?,?,?,?,?,?,?,?,?,?)" as String, allcols);
      }
   }   //end of sql.eachRow("select distinct seq_acc from seq_ipr_tb"


   //update other columns
   System.err.println "Update tf_tb table .....";
   sql.execute("CREATE INDEX if not exists ${tf_tb}_seq_acc_index ON ${tf_tb}(seq_acc)" as String);
   sql.execute("update ${tf_tb} set seq_realacc=(select distinct seq_realacc from ${seq_iprs_tb} where seq_acc=${tf_tb}.seq_acc)" as String);
   sql.execute("update ${tf_tb} set ipr_accs=(select distinct ipr_accs from ${seq_iprs_tb} where seq_acc=${tf_tb}.seq_acc)" as String);
   sql.execute("update ${tf_tb} set seq_comment=(select distinct seq_comment from ${seq_iprs_tb} where seq_acc=${tf_tb}.seq_acc)" as String);
   sql.execute("update ${tf_tb} set all_text=seq_acc || ' ' || seq_realacc || ' ' || ipr_accs || ' ' || seq_comment" as String);
   sql.execute("CREATE INDEX if not exists ${tf_tb}_all_text_index ON ${tf_tb}(all_text)" as String);
   sql.execute("CREATE INDEX if not exists ${tf_tb}_myscore_index ON ${tf_tb}(myscore)" as String);
   sql.execute("CREATE INDEX if not exists ${tf_tb}_ID_index ON ${tf_tb}(ID)" as String);

   //calculate order
   System.err.println "Calculate order column for each prediction based on score in tf_tb table .....";
   def seqrank=[:];
   sql.eachRow("select seq_acc,ID from ${tf_tb} order by seq_acc,myscore desc" as String){
      if(seqrank[it.seq_acc]==null)  seqrank[it.seq_acc]=1;
      else                           seqrank[it.seq_acc]++;
      sql.execute("update ${tf_tb} set rank=${seqrank[it.seq_acc]} where ID=${it.ID}" as String); 
   }

   def f = new File(tfTxtResultFullPath);
   f.append( ["family", "family_type", "seq_acc", "seq_realacc", "ipr_accs", "seq_comment"].join("\t") +"\n" );
   sql.eachRow("select * from ${tf_tb}" as String){
       f.append(  [it.family, it.family_type, it.seq_acc, it.seq_realacc, it.ipr_accs, it.seq_comment].join("\t") +"\n"  );
   }

   //close database
   sql.execute("COMMIT"); 
   sql.close();
 } //end of function

 public static def calPatternScore(def myexp){
   if(!(myexp=~/[1-9]/))  return 0.0;

   def pNumber="[0-9\\.eE+-]+";
   int i=0;
   int count=100;
   while( !myexp.isDouble()  &&  !myexp.isInteger() ) {
     i++;
     if(i>count) throw new Exception("Found error in calculated expression");
     myexp=myexp.replaceAll(/(${pNumber})&(${pNumber})/,{  full,first,second -> 
                                                           if(  (first as Double)==0.0 || (second as Double)==0.0 ) { 
                                                             full="0.0";
                                                           }else{
                                                             def nfirst=1-(first as Double);
                                                             def nsecond=1-(second as Double);
                                                             def noverlap=nfirst*nsecond;
                                                             def allnproba=(nfirst-noverlap)+(nsecond-noverlap)+noverlap; 
                                                             full=(1-allnproba) as String; 
                                                           }
                                                        }); //AND first
     myexp=myexp.replaceAll(/(${pNumber})\|(${pNumber})/,{  full,first,second -> 
                                                            full=(first as Double)>(second as Double)?first:second; 
                                                         });    //OR second
     myexp=myexp.replaceAll(/\((${pNumber})\)/,{full,first -> full=first});              //clean ( )
   }
   return myexp as Double;
 }

/*
 public static def predictTF(def tfdomainfile, def dbfile, def tfTxtResultFullPath, def families){
   def seq_ipr_tb=SessionProps.seq_ipr_tb;
   def seq_iprs_tb=SessionProps.seq_iprs_tb;
   def tf_tb=SessionProps.tf_tb;

   def sql = Sql.newInstance("jdbc:sqlite:"+dbfile, "", "", "org.sqlite.JDBC");
   sql.execute("create table if not exists ${tf_tb} (family text key, family_type text key, seq_acc text, seq_realacc text key, ipr_accs text, seq_comment text, all_text text key)" as String);
   sql.execute("PRAGMA journal_mode=OFF");
   sql.execute("PRAGMA synchronous=OFF");
   sql.execute("BEGIN TRANSACTION");

   new File(tfdomainfile).eachLine{
      if(it=~/^\s*$/) return;
      def cols=it.split("\t");
      if(cols.length!=3){
         System.err.println("Found error in ${tfdomainfile}, line: ${it}" as String);
         return;
      }
      def tffamily=cols[0];

      def musthave=(cols[1] as String).replaceAll("\\s+","").replaceAll("IPR"," ipr_accs like IPR").replaceAll("\\|"," or ").replaceAll("&"," and ").replaceAll(/IPR[T0-9]+/,'\'%$0%\'');
      def wheresql="("+musthave+")";

      def exclude=(cols[2] as String).replaceAll("\\s+","");        //ARF	(IPR010525|IPR011525)&IPR003340	-
      if(exclude=~/IPR[T0-9]+/){
          exclude=exclude.replaceAll("IPR"," ipr_accs like IPR").replaceAll("\\|"," or ").replaceAll("&"," and ").replaceAll(/IPR[T0-9]+/,'\'%$0%\'');
          exclude=" and not ("+exclude+")";
          wheresql=wheresql+exclude;
      }
      def f = new File(tfTxtResultFullPath);
      def querysql="select '${tffamily}' as family, seq_acc, seq_realacc, ipr_accs, seq_comment from ${seq_iprs_tb} where ${wheresql}" as String;
      sql.eachRow(querysql){
          f.append("${it.family}\t"+families[it.family].family_type+"\t${it.seq_acc}\t${it.seq_realacc}\t${it.ipr_accs}\t${it.seq_comment}\n");
          def all_text="${it.family} ${it.seq_acc} ${it.seq_realacc} ${it.ipr_accs} ${it.seq_comment}";
          sql.execute("insert into ${tf_tb} values(?,?,?,?,?,?,?)" as String, [it.family, families[it.family].family_type, it.seq_acc, it.seq_realacc, it.ipr_accs, it.seq_comment, all_text]);
      }
   }

   sql.execute("COMMIT"); 
   sql.close();
 } //end of function
*/

 public static def txt2db(def txtfile, def dbfile, def checked_query){
   def seq_ipr_tb=SessionProps.seq_ipr_tb;
   def seq_ipr_tb2=SessionProps.seq_ipr_tb+"2";
   def seq_iprs_tb=SessionProps.seq_iprs_tb;
   def sql = Sql.newInstance("jdbc:sqlite:"+dbfile, "", "", "org.sqlite.JDBC");
   
   //make seq_ipr_tb table
   try { sql.execute("drop table if exists ${seq_ipr_tb}" as String); } 
   catch(Exception e){}
   sql.execute("create table ${seq_ipr_tb} (seq_acc varchar(255), seq_realacc varchar(255), ipr_acc varchar(255), ipr_prog varchar(255), ipr_progacc varchar(255), seq_comment text, myscore real)" as String);
   sql.execute("PRAGMA journal_mode=OFF");
   sql.execute("PRAGMA synchronous=OFF");
   sql.execute("BEGIN TRANSACTION");
   def fasta = new Fasta(checked_query);
   new File(txtfile).eachLine{
     def tt = it.split("\t");  //S_155   IPR013729       PFAM    PF08523 4.4e-14 1       42      26      66
     def seq=fasta.getFullseq(tt[0]);
     def seq_realacc=seq.seqID;
     def seq_comment=seq.desc;
     try{ 
        def newIDseq=new SeqNewID(seq);
        seq_realacc=newIDseq.realSeqID;
        seq_comment=newIDseq.realDesc;
     }catch(Exception e) { }

     if(tt.length==12 && tt[10].isInteger() ) sql.execute("insert into ${seq_ipr_tb} values (?,?,?,?,?,?,?)" as String, [  tt[0], seq_realacc, tt[1], tt[2], tt[3], seq_comment, tt[11]  ]);
   }
   fasta.close();
   sql.execute("CREATE INDEX if not exists seq_acc1_index ON ${seq_ipr_tb}(seq_acc)" as String);
   sql.execute("CREATE INDEX if not exists seq_realacc1_index ON ${seq_ipr_tb}(seq_realacc)" as String);
   sql.execute("COMMIT"); 
  

   //make distinct seq_ipr_tb table: seq_ipr_tb2
   try { sql.execute("drop table if exists ${seq_ipr_tb2}" as String); } 
   catch(Exception e){}
   sql.execute("create table ${seq_ipr_tb2} (seq_acc varchar(255), seq_realacc varchar(255), ipr_acc varchar(255), seq_comment text, myscore real)" as String);
   sql.execute("BEGIN TRANSACTION");
   sql.execute("insert into ${seq_ipr_tb2} select distinct seq_acc,seq_realacc,ipr_acc,seq_comment,myscore from ${seq_ipr_tb}" as String);
   sql.execute("CREATE INDEX if not exists seq_acc2_index ON ${seq_ipr_tb2}(seq_acc)" as String);
   sql.execute("CREATE INDEX if not exists seq_realacc2_index ON ${seq_ipr_tb2}(seq_realacc)" as String);
   sql.execute("COMMIT"); 


   //------make seq_iprs_tb from seq_ipr_tb2
   try { sql.execute("drop table if exists ${seq_iprs_tb}" as String); } 
   catch(Exception e){}
   sql.execute("create table ${seq_iprs_tb} (seq_acc varchar(255), seq_realacc varchar(255), ipr_accs text, seq_comment text)" as String);
   sql.execute("BEGIN TRANSACTION");
   sql.execute("insert into ${seq_iprs_tb} select seq_acc,seq_realacc,group_concat(ipr_acc),seq_comment from ${seq_ipr_tb2} group by seq_acc,seq_realacc" as String);
   sql.execute("CREATE INDEX if not exists seq_acc3_index ON ${seq_iprs_tb}(seq_acc)" as String);
   sql.execute("CREATE INDEX if not exists seq_realacc3_index ON ${seq_iprs_tb}(seq_realacc)" as String);
   sql.execute("CREATE INDEX if not exists ipr_accs3_index ON ${seq_iprs_tb}(ipr_accs)" as String);
   sql.execute("COMMIT"); 
   
   sql.close();
 } //end of function

} //end of class
