#!/usr/bin/perl  -w


use strict;
use DBI();
use Getopt::Long;
use Scalar::Util qw(looks_like_number);

unshift( @INC, $ENV{'BINDIR'} ) if( defined $ENV{'BINDIR'} );

my $myprompt = '
This script calculate the comprehensive score of miniIPR prediction

Usage:

    calscore.pl <Installation folder of miniIPR> <output of miniIPR>

';

my $help;
my $myresult = GetOptions ("h" => \$help);

if( !defined $myresult || ! defined $ARGV[1]  || defined $help)
{
    print  $myprompt;  exit 1;
}

my $probafn  = "${ARGV[0]}/data/probatb.sqlite";
my $weightfn = "${ARGV[0]}/data/pweight";
my $iprlistfn = "${ARGV[0]}/data/iprlist.txt";
my $miniIPRoutput = $ARGV[1];

my %weight;
open(MYFILE, $weightfn) || die "Failed to open subprogram weight table file name!\n";
while(my $line=<MYFILE>){
  chop $line;
  my @cols=split(/\t/,$line);
  next if(!defined $cols[1]||$cols[1]=~/^\s*$/||$cols[0]=~/^\s*$/);
  $weight{$cols[0]}=$cols[1];
}
close(MYFILE);


my %validipr;
open(MYFILE,$iprlistfn) || die "Failed to open $iprlistfn";
while(my $line=<MYFILE>)
{
  chop $line;  $line=~s/^\s*|\s*$//g;  next if($line eq "");  
  $validipr{$line}=1;
}
close(MYFILE);


my %finalproba;
my $dbh = DBI->connect("dbi:SQLite:dbname=${probafn}","","") || die "Failed to open SQLite db ${probafn}.\n";
open(FILE2,$miniIPRoutput) || die "Failed to open miniIPR output file ${miniIPRoutput}";
open(FILE3,">${miniIPRoutput}.2") || die "Failed to open ${miniIPRoutput}.2 output file";
my $tmpseqacc=""; my $tmpiprdomain=""; my $tmpprogram="";  my $tmpevalue="";
while(my $line=<FILE2>){
  #AT1G75240.1     IPR006456       HMMPfam PF04770 3.2e-34 2       58      72      127
  #AT1G14920.1     IPR005202       HMMPfam PF03514 1.4e-136        1       374     169     529
  chop $line; $line=~s/^\s*|\s*$//g;  next if($line eq "");
  my @cols=split(/\t/,$line);
  if(@cols<9){
    print STDERR "Found error in miniIPR output file: $line\n";
    next;
  }

  next if( !defined $validipr{$cols[1]} );

  if($tmpseqacc eq $cols[0] && $tmpiprdomain eq $cols[1] && $tmpprogram eq $cols[2]) {  # multiply HSP for query/domain
    print FILE3 $line, "\tNA\tNA\n";
    next;
  }
  my $myproba;
  $tmpseqacc=$cols[0]; $tmpiprdomain=$cols[1]; $tmpprogram=$cols[2];   $tmpevalue=$cols[4];
  if(!looks_like_number($tmpevalue)){
    $myproba=1.0;
  }else{
    my $sth = $dbh->prepare("select myproba,evalue from probatb where program='${tmpprogram}' and evalue<=${tmpevalue} order by evalue desc limit 0,1");
    $sth->execute() ||  die "Failed to access scoring matrix table db!\n";
    my $proba1; my $evalue1;
    ($proba1, $evalue1) = $sth->fetchrow;
    if(!defined $proba1) {         $myproba=1.0;      }
    elsif($tmpevalue==$evalue1){   $myproba=$proba1;  }
    else{
       $myproba=$proba1;
       #------
       #my $sth2 = $dbh->prepare("select myproba,evalue from probatb where program='${tmpprogram}' and evalue>${tmpevalue} order by evalue asc limit 0,1");
       #$sth2->execute() ||  die "Failed to access scoring matrix table db!\n";
       #my $proba2; my $evalue2;
       #($proba2, $evalue2) = $sth2->fetchrow;
       #$sth2->finish;
       #------
       #if(defined $proba2){
       #    my $myratio=abs($tmpevalue-$evalue1)/abs($evalue2-$evalue1);
       #    my $proba_dif=abs($proba1-$proba2)*$myratio;
       #    if($myratio<1.0){
       #       $myproba=$myproba-$proba_dif;
       #       print STDERR "For ${tmpseqacc}|${tmpiprdomain}|${tmpprogram}|${tmpevalue}: Adjust proba ${proba1} ->", $myproba , \
       #           "    ratio=${myratio}; proba_dif=${proba_dif} (${proba2}/${evalue2}<=>${myproba}/${tmpevalue}<==>${proba1}/${evalue1})\n";
       #    }
       #}
    }
    $sth->finish;
  }
  my $tmpweight=$weight{$tmpiprdomain};
  my $tmpkey="${tmpseqacc}\t${tmpiprdomain}";
  $finalproba{$tmpkey}=0.0 if(!defined $finalproba{$tmpkey});
  $finalproba{$tmpkey}=$finalproba{$tmpkey}+$myproba/$tmpweight;
  print FILE3 $line,"\t",$myproba,"\t",$tmpweight,"\n";
}
close(FILE3);
close(FILE2);
$dbh->disconnect;

#------print score with miniIPR raw output
open(FILE2,"${miniIPRoutput}.2") || die "Failed to open miniIPR output file";
while(my $line=<FILE2>){
  #AT1G75240.1     IPR006456       HMMPfam PF04770 3.2e-34 2       58      72      127
  chop $line; $line=~s/^\s*|\s*$//g;  next if($line eq "");
  my @cols=split(/\t/,$line);
  next if(@cols<9);

  my $tmpseqacc=$cols[0]; my $tmpiprdomain=$cols[1];
  my $tmpkey="${tmpseqacc}\t${tmpiprdomain}";
  print $line;
  if(defined $finalproba{$tmpkey}) { print "\t", $finalproba{$tmpkey},"\n";  }
  else                             { print "\tNA\n";  }

}
close(FILE2);
`rm -f ${miniIPRoutput}.2`;


sub log10 {
   my $n = shift;
   return log($n)/log(10);
}

