=head1 NAME

PfamPostFilter - Pfam postfiltering utilities.

=head1 SYNOPSIS

  # examples here
  use Dispatcher::Tool::Hmmer::PfamPostFilter;

=head1 DESCRIPTION

=head1 VERSIONS

$Id: PfamPostFilter.pm,v 1.15 2011/03/24 11:30:36 hunter Exp $

Copyright (c) European Bioinformatics Institute 2005

=head1 AUTHORS / ACKNOWLEDGEMENTS

Sarah Hunter <hunter@ebi.ac.uk>
Emmanuel Quevillon <tuco@ebi.ac.uk>

=cut


package Dispatcher::Tool::Hmmer::PfamPostFilter;


BEGIN {
    unless($ENV{IPRSCAN_HOME}){
	print STDERR "$0 : IPRSCAN_HOME environment variable not set!\n";
	exit();
    }
    unless($ENV{IPRSCAN_LIB}){
	$ENV{IPRSCAN_LIB} = "$ENV{IPRSCAN_HOME}/lib";
    }
}

use lib $ENV{IPRSCAN_LIB};
use Dispatcher::Tool::Hmmer3;
@ISA = qw( Dispatcher::Tool::Hmmer3 );
use Data::Dumper;
use English;
use strict;


=head2 new

 Description: Constructor allocates an anonymous hash, which
              is tied to the class.
 Arguments:   none
 Returns:     Blessed reference

=cut

sub new (){

    my($class) = @_;

    my ($self) = { };

    bless($self, $class);

    return $self;

}

=head2 isOverlapping

 Description:   Determines whether or not 2 hits overlap on the sequence
 Arguments:     hash ref, id1 id2
 Returns:       1, id if overlap
                1, '' if no overlap
                0, msg if error
=cut
                                                                                                                                                             
sub isOverlapping($$$) {

	# Need to simplify this overlap calculation eventually:
	# Phil's code:
        # return  !  ((one.getStart() > two.getStop()) ||
        # (two.getStart() > one.getStop()));

        my ($self, $hash, $id1, $id2) = @_;
        return (0, "isOverlapping: need results hash, plus 2 key ids as arguments.") unless ($hash, $id1, $id2);
                                                                                                                                                             
        my($s1, $e1, $s2, $e2) = ($hash->{$id1}->{start}, $hash->{$id1}->{end}, $hash->{$id2}->{start}, $hash->{$id2}->{end});
                                                                                                                                                             
        # Rule is, if we have an overlap for two Pfam hits, we need to check if
        # a) they are different Pfams (but one is a nested domain) => Apply nested domain filtering;
	# b) they are different Pfams (but the same clan) => Apply clan filtering;
                                                                                                                                                             
        #No overlap
        if(($e1 < $s2 || $e2 < $s1) || ($e1 == $s2 && $s1 < $s2) || ($e2 == $s1 && $s2 < $s1)){
                return (1, '');
        }
                                                                                                                                                             
        #  |===========================|        hit 1   #          |=============|   hit 1
        #                  |==================| hit 2   #   |===============|        hit 2
                                                                                                                                                             
        if(($s1 < $s2 && $e1 > $s2) || ($s2 < $s1 && $e2 > $s1)){
                return (1, $id1);
        }
                                                                                                                                                             
        #   |=========================|  hit 1    #          |=========|           hit 1
        #          |=========|           hit 2    #   |=========================|  hit 2
                                                                                                                                                             
        elsif(($s1 > $s2 && $e1 < $e2) || ($s2 > $s1 && $e2 < $e1)){
                return (1, $id1);
        }
                                                                                                                                                             
        #   |=========================|  hit 1    #   |=========|                  hit 1
        #   |=========|                  hit 2    #   |=========================|  hit 2
        #
        #   |=========================|  hit 1    #                   |=========|  hit 1
        #                   |=========|  hit 2    #   |=========================|  hit 2
                                                                                                                                                             
        elsif(($s1 == $s2 && ($e1 > $e2 || $e1 < $e2)) || ($e1 == $e2 && ($s1 < $s2 || $s1 > $s2))){
                return (1, $id1);
        }
                                                                                                                                                             
        #We never know, be prepared
        #   |=========================|  hit 1
        #   |=========================|  hit 2
        elsif($s1 == $s2 && $e1 == $e2){
                return (1, $id1);
        }
                                                                                                                                                             
return (0, "isOverlapping: Cannot determine overlap!");
                                                                                                                                                             
}

=head2 doLSvsFS
                                                                                                                                                             
 Description:   checks whether 2 ids are the same Pfam but different models and overlap
                applies appropriate filtering
		OBSOLETE in HMMER3
 Arguments:     hash ref of results
 Returns:       1, $hash modified hash with status set to "F" according to filtering
                0, msg if error
=cut
                                                                                                                                                             
sub doLSvsFS ($) {
                                                                                                                                                             
    my ($self, $hash) = @_;
    return (0, "doLSvsFS: Need hashref as an argument" ) unless ($hash && ref($hash) eq 'HASH');
    return (1, $hash) unless (keys %$hash);
														     
    my ($res, $msg);
														     
    my $ndomains = $hash->{ndomains};
														     
    ILOOP: for (my $i=1; $i<=$ndomains; $i++) {
	if ($hash->{$i}->{status} eq "F") {
	    next ILOOP;
	}

	JLOOP: for (my $j=$i+1; $j <=$ndomains; $j++) {
	    if ($hash->{$j}->{status} eq "F") {
		next JLOOP;
	    }

	    my($fullid1, $fullid2) = ($hash->{$i}->{methid}, $hash->{$j}->{methid});
	    my ($res, $pfid1) = $self->filterID($fullid1, 'hmmpfam');
	    return(0, "doLSvsFS: $pfid1") unless($res);
						     
	    my ($res, $pfid2) = $self->filterID($fullid2, 'hmmpfam');
	    return(0, "doLSvsFS: $pfid2") unless($res);
						     
	    if ($pfid1 eq $pfid2) { # LS and FS hit from the same Pfam entry
		my ($res, $overlapping) = $self->isOverlapping($hash,$i,$j);
		return (0, "doLSvsFS: $overlapping") unless $res;
		if ($overlapping) {
		    ($res, $msg) = $self->applyAmFilter($hash, $i, $j);
		    return (0, "doLSvsFS: Cannot apply AM filter: $msg") unless ($res);
		    $hash->{$msg}->{status}="F" if $res;
		}
	    }
	}
    }

    return (1, $hash);
															 
}

=head2 removeOverlap

 Description: Used to identify when 2 Pfam ids are overlapping.
              If they are, check they have a significant sequence score for that domain
              Check if they are nested domains,
              Or members of the same clan
              And remove according to each rule
 Arguments:   $hash  Hash table containing hits for a protein
 Returns:     1, modified $hash (with overlaps removed)
              0, $msg on error

=cut

sub removeOverlap($) {

    my($self, $hash) = @_;

    return(0, "removeOverlap: Need results hash table as an argument.") unless($hash);
    return(0, "removeOverlap: Need a reference to a hash table.") unless(ref($hash) eq 'HASH');
    return(1, $hash) unless(keys %$hash);

    my @remove = (); # For storing Key IDs to be removed
    my %safe; # For storing Key IDs that are "safe" (e.g. nested domains)
    my ($res,$msg);
                                                                                                                                                             
    # Find out how many pfams are hitting
    my $ndomains = $hash->{ndomains};
                                                                                                                                                             
    # Go through them, one by one and pair-wise compare them (but not to self)
    for(my $i = 1 ; $i <= $ndomains; $i++) {
                                                                                                                                                             
        # But first of all, swap the long description to the short name by doing a lookup in Pfam-A
        ($res,$msg) = $self->lookupShortName($hash,$i);
        return(0,"removeOverlap:lookup of shortname failed: $msg") unless ($res);
        $hash->{$i}->{desc}=$msg;
                                                                                                                                                             
        my $seqscore1 = $hash->{$i}->{seqscore}; # check if there is a significant sequence score for that domain
        unless ($seqscore1){            # if not, push onto the remove list and skip to the next hit
            push (@remove,$i);
            $hash->{$i}->{status}="F";
            next;
        }
                                                                                                                                                             
        for(my $j = $i+1; $j <= $ndomains; $j++) {

            my $seqscore2 = $hash->{$j}->{seqscore};# check if there is a significant sequence score for that domain
            unless ($seqscore2) {       # if not, push onto the remove list and skip to the next hit
                $hash->{$j}->{status}="F";
                push (@remove,$j);      
                next;
            }

            my($fullid1, $fullid2) = ($hash->{$i}->{methid}, $hash->{$j}->{methid});
            my ($res, $pfid1) = $self->filterID($fullid1, 'hmmpfam');
            return(0, "removeOverlap: $pfid1") unless($res);

            my ($res, $pfid2) = $self->filterID($fullid2, 'hmmpfam');
            return(0, "removeOverlap: $pfid2") unless($res);

            my ($res, $overlapping) = $self->isOverlapping($hash,$i,$j);
            return (0, "removeOverlap: $overlapping") unless ($res);

            if ($overlapping) {

                #We found an overlap and so a hit needs to be removed. But we need to check which, according to the
                #various rules (Clan, Nested Domain)
                #See the Pfam website for information about the above rules.
                                                                                                                                                             
                # Check 1) if nested domain (in which case, keep) or 2) in same clan

                #If one Domain is nested in another, because we are going thru the hits sequentially,
                #domain $j must be the one that is nested, if at all, because its start is later in the
                #sequence than $i's

                # Is $j nested in $i?
                ($res,$msg) = $self->hasNestedDomain($hash,$i,$j);
                return(0, "removeOverlap: $msg") unless $res;
                $safe{$msg} = "" if ($msg);
                next if ($msg);
                                                                                                                                                             
		# Are $j and $i from the same clan?
                ($res,$msg) = $self->areFromSameClan($hash,$i,$j);
                return(0, "removeOverlap: $msg") unless $res;
                push(@remove,$msg) if ($msg);
		$hash->{$msg}->{status}="F" if ($msg);
                next if ($msg);
                                                                                                                                                             
            } # End of ($overlapping) loop
        } # End of $j loop
    } # End of $i loop
                                                                                                                                                             
    # Go through Key IDs earmarked for deletion and remove them from the hash
    for(@remove) {
        delete ($hash->{$_}) if ($_ =~/^\d+$/ && !$safe{$_});
    }
                                                                                                                                                             
    return(1, $hash);
}

=head2 areFromSameClan

    Description: Get Pfam entries and check if they belong to the same clan.
    Arguments:   $hash Reference to a hash table containing parsed results from Pfam
                 $id1 Key number id of the domains being compared
                 $id2 Key number id of the domains being compared
    Returns:     1, id to be removed
                 0  '' on error

=cut

sub areFromSameClan($$$){

    require Index::PfamClan;

    my ($self, $hash, $id1, $id2) = @_;

    return(0, "areFromSameClan: no list given") unless($hash && $id1 && $id2);
    return(0, "areFromSameClan: need a reference to a hash table") unless(ref($hash) eq 'HASH');

    my ($res, $msg, $pfclan, $clan1, $clan2, $entry1, $entry2);
    my ($fullid1,$fullid2) = ($hash->{$id1}->{methid}, $hash->{$id2}->{methid});
    my ($score1,$score2) = ($hash->{$id1}->{score}, $hash->{$id2}->{score});

    ($res, $pfclan) = Index::PfamClan->new(undef, 1);
    return(0, $pfclan) unless $res;

    my ($res, $pfid1) = $self->filterID($fullid1, 'hmmpfam');
    return(0, "areFromSameClan: $fullid1") unless($res);
                                                                                                                                                             
    my ($res, $pfid2) = $self->filterID($fullid2, 'hmmpfam');
    return(0, "areFromSameClan: $fullid2") unless($res);
                                                                                                                                                             
    ($res, $entry1) = $pfclan->getEntry($pfid1);
    $entry1 = $entry1->[0] if($entry1);
    return(0, "areFromSameClan: $entry1") unless $res;
                                                                                                                                                             
    ($res, $entry2) = $pfclan->getEntry($pfid2);
    $entry2 = $entry2->[0] if($entry2);
    return(0, "areFromSameClan: $entry2") unless $res;
                                                                                                                                                             
    #One or more of these Pfam entries do not belong to a clan so clan filtering
    #cannot apply
    return(1,'')  unless($entry1 && $entry2);
                                                                                                                                                             
    ($res, $clan1) = $pfclan->getField('clan',\$entry1);
    $clan1 = $clan1->[0] if($clan1);
    return(0, "areFromSameClan: $clan1") unless $res;
                                                                                                                                                             
    ($res, $clan2) = $pfclan->getField('clan',\$entry2);
    $clan2 = $clan2->[0] if($clan2);
    return(0, "areFromSameClan: $clan2") unless $res;
                                                                                                                                                             
    if($clan1 eq $clan2){
        # They are the same clan, so check which has the better score and discard the other one
        # But if either of them are false, don't bother with a comparison - the false hit will be removed anyway
        if ($hash->{$id1}->{status} ne "F" && $hash->{$id2}->{status} ne "F") {
            if($score1<=$score2) {
                return (1, $id1);
            } else {
                return (1, $id2);
            }
        } else { # else for checking false status
            return(1,'');
        }
    } else { # else for checking if in same clan
        return(1, '');
    }
}

=head2 hasNestedDomain
                                                                                                                                                             
    Description: Check if two overlapping Pfam hits are related by a nested rule.
                 i.e. is the second id (argument) nested in the first?
                 We assume that the hits are processed in the order they appear on the
                 sequence, so that any nested domain would have to be the second one of the pair
                                                                                                                                                             
    Arguments:   $hash  reference to results hash
                 $id1  Key domain number ID
                 $id2  Key domain number ID
                                                                                                                                                             
    Returns:     1, '$id2'  if it is nested (should be kept in "safe" list)
                 1, '' if is not nested
                                                                                                                                                             
=cut
                                                                                                                                                             
sub hasNestedDomain($$$) {
                                                                                                                                                             
    require Index::PfamA;
                                                                                                                                                             
    my($self, $hash, $id1, $id2) = @_;
    return(0, "hasNestedDomain: Need 2 ids and hash reference.") unless($hash && $id1 && $id2);
    return(0, "need a hash reference") unless(ref($hash) eq 'HASH');
                                                                                                                                                             
    my($res, $pfama) = Index::PfamA->new(undef, 1);
    return(0, "hasNestedDomain: $pfama") unless($res);

    if ($hash->{$id1}->{status} eq "F") { # The "parent" hit is not valid, so ignore this comparison
        return (1,'');
    }
    
    my($fullid1,$fullid2) = ($hash->{$id1}->{methid},$hash->{$id2}->{methid});
    my ($nested,$entry);
    my $methid1;
                                                                                                                                                             
    # Clean up the first ID so that no model type on the end (if it exists)
    if ($fullid1 =~ /(\w+\.\d+)\.[lf]s/) {
        $methid1 = $1;
    } else {
        $methid1 = $fullid1;
    }
                                                                                                                                                             
    # Clean up the second ID so no version number or model type on the end
    my ($res, $pfid2) = $self->filterID($fullid2, 'hmmpfam');
    return(0, "hasNestedDomain: $fullid2") unless($res);
                                                                                                                                                             
    # Check second ID and see if it is nested in the first (purely based on if #=GF NE field in
    # the Pfam-A file is methid2, and the entry has methid1 as accession.)
    # NE is the domain that is nested within the entry
                                                                                                                                                             
    # retrieve the entry you wish to check for the nested domain (i.e. methid1's entry)
    ($res, $entry) = $pfama->getEntry($methid1);
    $entry = $entry->[0] if $entry;
    return(0, "hasNestedDomain: $entry") unless ($res);
                                                                                                                                                             
    # retrieve the field from the entry
    ($res, $nested) = $pfama->getField('nested',\$entry);
    $nested = $nested->[0] if $nested;
    return(0, "hasNestedDomain: $nested") unless($res);
                                                                                                                                                             
    if ($nested eq $pfid2) {
        return(1,$id2);
    } else {
        return(1,'');
    }
}
                                                                                                                                                             
=head2 applyModFilter
                                                                                                                                                             
        Description:    Checks if both ls and fs models hit and picks the ls model preferentially
			N.B. Replaced by applyAmFilter in v19 of Pfam (02-feb-2006)
        Arguments:      $hash Reference to a hash table containing parsed results from Pfam
                        $id1 Key number id of the domain being compared
                        $id2 Key number id of the domain being compared
        Returns:        1, $id to be removed
                        0 '' on error
=cut
                                                                                                                                                             
sub applyModFilter($$$) {
                                                                                                                                                             
        my ($self, $hash, $id1, $id2) = @_;
                                                                                                                                                             
        return (0, "applyModFilter: need 3 arguments") unless ($hash, $id1, $id2);
        return (0, "applyModFilter: \$hash not a hashref") unless (ref($hash) eq 'HASH');
                                                                                                                                                             
        my ($res, $msg);
                                                                                                                                                             
        my $fullid = $hash->{$id1}->{methid};
                                                                                                                                                             
        # Only need to check one of the hits, as if not it, then will be other that is discarded
        $fullid=~/^\w+\.\d+\.([fl]s)/;
        my $model=$1;
                                                                                                                                                             
        if ($model eq "ls") {
                return(1,$id2);
        } else {
                return(1,$id1);
        }
}
                                                                                                                                                             
=head2 applyAmFilter
                                                                                                                                                             
        Description:    Checks if both ls and fs model hit and picks according to Pfam-A AM field
        Arguments:      $hash Reference to a hash table containing parsed results from Pfam
                        $id1 Key number id of the domains being compared
                        $id2 Key number id of the domains being compared
        Returns:        1, id to be removed
                        0 '' on error
                                                                                                                                                             
=cut
                                                                                                                                                             
sub applyAmFilter($$$) {
                                                                                                                                                             
        require Index::PfamA;
                                                                                                                                                             
        my ($self, $hash, $id1, $id2) = @_;
                                                                                                                                                             
        return(0, "applyAmFilter: need 3 arguments") unless($hash,$id1,$id2);
        return(0, "applyAmFilter: \$hash not a hashref") unless(ref($hash) eq 'HASH');
                                                                                                                                                             
        my ($res, $msg, $pfseed, $amfilter, $entry);
        my $fullid = $hash->{$id1}->{methid};
                                                                                                                                                             
        # Only need to check one of the hits, as if not it, then will be other that is discarded
        $fullid=~/(^\w+\.\d+)\.([fl]s)/;
        my ($methid,$model)=($1,$2);
                                                                                                                                                             
        ($res, $pfseed) = Index::PfamA->new(undef, 1);
        return(0, $pfseed) unless $res;
                                                                                                                                                             
        ($res, $entry) = $pfseed->getEntry($methid);
        return(0, "applyAmFilter: entry retrieval failure for $entry") unless $res;
        $entry = $entry->[0] if ($entry);
	return (0, "applyAmFilter: empty seed entry for $methid, check file versions are the same for HMM and SEED files") if ($entry eq "");
                                                                                                                                                             
        ($res, $amfilter) = $pfseed->getField('am',\$entry);
        return(0, "applyAmFilter: am retrieval failure for $amfilter") unless $res;
        $amfilter = $amfilter->[0] if ($amfilter);
                                                                                                                                                             
                                                                                                                                                             
        if ($amfilter =~ /globalfirst/) {
                #discard FS
                if ($model eq "fs") {
                        return(1,$id1);
                }else {
                        return(1,$id2);
                }
        } elsif ($amfilter =~ /localfirst/) {
                #discard LS
                if ($model eq "ls") {
                        return(1,$id1);
                } else {
                        return(1,$id2);
                }
        } elsif ($amfilter =~ /byscore/) {
                # pick model with best e-value (i.e. lowest)
		my($score1, $score2) = ($hash->{$id1}->{evalue}, $hash->{$id2}->{evalue}); # byscore actually means "by e-value"!
                if ($score1 > $score2) {
                        return(1,$id1);
                } elsif ($score1 < $score2) {
                        return(1,$id2);
		} elsif ($score1 == $score2) { # might happen, rarely!
			# if e-values are equal, always choose the LS and discard FS hit
			if ($model eq "fs") {
				return(1,$id1);
			} else {
				return(1,$id2);
			}
                }
        } else {
                #shouldn't get this far
                return (0, "$msg: error!");
        }
}

=head2 lookupShortName
                                                                                                                                                             
        Description:    Replaces the long description from Pfam with the short name (i.e. ID)
        Arguments:      $hash Reference to a hash table containing parsed results from Pfam
                        $id1 key id for the current hit being changed
        Returns:        1, modified description
                        0 '' on error
=cut
                                                                                                                                                             
sub lookupShortName($) {
                                                                                                                                                             
    my ($self,$hash,$id1) = @_;
    return(0, "lookupShortName: Need results hash table and id as an argument.") unless($hash && $id1);
    return(0, "lookupShortName: Need a reference to a hash table.") unless(ref($hash) eq 'HASH');
                                                                                                                                                             
    require Index::PfamA;
                                                                                                                                                             
    my($res, $pfama) = Index::PfamA->new(undef, 1);
    return(0, "lookupShortName: $pfama") unless($res);
                                                                                                                                                             
    my $fullid1 = $hash->{$id1}->{methid};
    my ($shortname,$entry);
    my $methid1;
                                                                                                                                                             
    # Clean up the first ID so that no model type on the end (if it exists)
    if ($fullid1 =~ /(\w+\.\d+)\.[lf]s/) {
        $methid1 = $1;
    } else {
        $methid1 = $fullid1;
    }
                                                                                                                                                             
    # retrieve the entry you wish to check for the nested domain (i.e. methid1's entry)
    ($res, $entry) = $pfama->getEntry($methid1);
    $entry = $entry->[0] if $entry;
    return(0, "lookupShortName: $entry") unless ($res);
                                                                                                                                                             
    # retrieve the field from the entry
    ($res, $shortname) = $pfama->getField('id',\$entry);
    $shortname = $shortname ->[0] if $shortname;
    return(0, "lookupShortName: $shortname") unless($res);
                                                                                                                                                             
    return(1,$shortname);
                                                                                                                                                             
}

