#!/usr/bin/perl
use strict;
# fasta_checker.pl
# http://supfam.org
# Copyright (c) 2001 MRC and Julian Gough; see http://supfam.org/SUPERFAMILY/license.html
#This program is supposed to: Check the sanity of a fasta sequence file, outputting the
#corrected sequences to standard out,  and reporting to standard error.
#Julian Gough 22.1.01

#ASIGN-VARIABLES-------------------------------------

my $input;
my $usage;
my $line=0;
my %ids;
my @seqline;
my $residue;
my $comment;
my $flag=0;
my $idflag=0;
my $thisline;
my $verbose;
my $one;
my $two;

#----------------------------------------------------


#ARGUEMENTS-------------------------------------------
$usage="check_fasta.pl  <file.fa> <verbosity #>";
die "Usage: $usage\n" unless (@ARGV > 0);

$input=$ARGV[0];
$verbose=$ARGV[1];
unless (defined($verbose)){
  $verbose=0;
}
#-----------------------------------------------------


#CHECK-IT---------------------------------------------
open FILE,("$input");
while (<FILE>){
  $line++;
  if (/^>/){
    #Is there any sequence for the ID?
    if($flag == 0 and $line > 1){
      print STDERR "No Sequence,  line:",($line-1),"\n";
      if ($verbose == 1){
	print STDERR $_;
      }
    }
    if (/^>\s+(\S+)$/ or /^>\s+(\S+)(\s.*)$/){
      print STDERR "Space before identifier,  line:$line\n";
      if ($verbose == 1){
	print STDERR $_;
      }
      if (defined($2)){
	$thisline=">$1$2\n";
      }
      else{
	$thisline=">$1\n";
      }
    }
    else{
      $thisline=$_;
    }
    if ($thisline =~ /^(>\S+)$/ or $thisline =~ /^(>\S+)(\s.*)$/){
      #Is ID unique?
      if (exists($ids{"$1"})){
	print STDERR "ID $1 not unique,  line:$line\n";
	if ($verbose == 1){
	  print STDERR $_;
	}
	$idflag=1;
      }
      else{
	$idflag=0;
	$ids{"$1"}=1;
	#Does the ID end in a comma?
	$one =$1;
	$two=$2;
	if ($one =~ /^(\S+)\,$/){
	  $comment = "$1$two\n";
	  print STDERR "ID $one ends in a comma,  line:$line\n";
	}
	else{
	  $comment = $thisline;
	}
	$flag=0;
      }
    }
    #Identifier present?
    else{
      print STDERR "No identifier,  line:$line\n";
      if ($verbose == 1){
	print STDERR $_;
      }
    }
  }
  elsif ($idflag == 0){
    if (/\S/){
      @seqline = split //,$_;
      foreach $residue (@seqline){
	if ($residue =~ /\w/){
	  #Not an amino acid?
	  if (uc($residue) eq 'J' or uc($residue) eq 'O' or $residue =~ /\d/){
	    print STDERR "Not an amino acid, '$residue',  line:$line\n";
	    if ($verbose == 1){
	      print STDERR $_;
	    }
	  }
	  else{
	    #Is there any sequence for the ID?
	    if($flag == 0){
	      $flag=1;
	      print $comment;
	    }
	    print $residue;
	  }
	}
	#Not a letter?
	else{
	  if ($residue eq "\n"){
	    print $residue;
	  }
	  #Blank?
	  elsif ($residue eq ' '){
	    print STDERR "Blank sequence character,  line:$line\n";
	    if ($verbose == 1){
	      print STDERR $_;
	    }
	  }
	  else{
	    print STDERR "Character '$residue',  line:$line\n";
	    if ($verbose == 1){
	      print STDERR $_;
	    }
	  }
	}
      }
    }
    #Blank line?
    else{
      print STDERR "Blank line,  line:$line\n";
    }
  }
}
close FILE;
#-----------------------------------------------------
